local gl = require("opengl")
local GL = gl

local Image = require("image")
local Array = require("Array")

local sketch = require("opengl.sketch")
local Texture = require("opengl.Texture")
local Shader = require("opengl.Shader")
local Mesh = require("opengl.Mesh")

local ctx = "OpenGL Template"
local win = Window(ctx, 1000, 0, 512, 512)
win.sync = true

local blobImg = Image(LuaAV.findfile("blob.png"))
local cutImg = Image(LuaAV.findfile("lines.png"))

local blobTex = Texture(ctx)
blobTex:fromarray(blobImg:array())
blobTex.target = GL.TEXTURE_2D

local cutTex = Texture(ctx)
cutTex:fromarray(cutImg:array())
cutTex.target = GL.TEXTURE_2D
cutTex.wrap = GL.REPEAT

local frameTex = Texture{
	ctx = ctx,
	target = GL.TEXTURE_2D,
	wrap = GL.CLAMP_TO_EDGE,
}

local psShader = Shader{
	ctx = ctx,
	file = LuaAV.findfile("ps.billboard.shl"),
	param = {
		spriteTex = 1,
		pointSize = 25,
	}
}

local lsShader = Shader{
	ctx = ctx,
	file = LuaAV.findfile("light.scatter.shl"),
	param = {
		frameTex = 0,
		lightPos = {0.5, 0.5},
		density = 0.9,
		weight = 0.9,
		decay = 0.96,
		exposure = 0.32,
		
	}
}

local dim = 60

local vertices = Array{
	components = 3,
	type = Array.Float32,
	dim = {dim * dim},
}

local normals = Array{
	components = 3,
	type = Array.Float32,
	dim = {dim * dim},
}

local texcoords = Array{
	components = 2,
	type = Array.Float32,
	dim = {dim * dim},
}

local indices = Array{
	components = 1,
	type = Array.UInt32,
	dim = {(dim-1) * (dim-1) * 4},
}

local sphereMesh = Mesh{
	ctx = ctx,
	primitive = GL.QUADS,
}

function genMesh(r, lats, longs)

	for i = 0, longs - 1 do
    
		local lng = 2 * math.pi * i / (longs - 1)
		local x = r * math.cos(lng)
		local y = r * math.sin(lng)
    
		for j = 0, lats - 1 do
			
			local lat0 = math.pi * (-0.5 + j / (lats - 1))
	
			local z0  = r * math.sin(lat0)
			local zr0 =  math.cos(lat0)

 			local normal = {x = x * zr0, y = y * zr0, z = z0 }
 			local len = math.sqrt(
 				normal.x * normal.x +
 				normal.y * normal.y +
 				normal.z * normal.z
 			)

			texcoords:setcell(j + i * dim, { (x * zr0) / 2 + 0.5, (y * zr0) / 1.8 + 0.5 })
			normals:setcell(j + i * dim, { normal.x / len, normal.y / len, normal.z / len })
			vertices:setcell(j + i * dim, { x * zr0, y * zr0, z0 })

		end
	end
	
	local index = 0

	for i = 0, dim-2 do
		for j = 0, dim-2 do

			indices:setcell(index, (j % dim) + (i % dim) * dim)
			indices:setcell(index + 1, (j % dim) + ((i + 1) % dim) * dim)
			indices:setcell(index + 2, ((j + 1) % dim) + ((i + 1) % dim) * dim)
			indices:setcell(index + 3, ((j + 1) % dim) + (i % dim) * dim)
			index = index + 4

		end
	end
end

function drawSphere(r, lats, longs)

	for i = 1, lats do
		local lat0 = math.pi * (-0.5 + (i - 1) / lats)
	
		local z0  = math.sin(lat0)
		local zr0 =  math.cos(lat0)
    
		local lat1 = math.pi * (-0.5 + i / lats)
		local z1 = math.sin(lat1)
		local zr1 = math.cos(lat1)
    
		gl.Begin(GL.QUAD_STRIP)
	
		for j = 0, longs do
			local lng = 2 * math.pi * (j - 1) / longs
 			local x = r * math.cos(lng)
 			local y = r * math.sin(lng)

			gl.TexCoord((x * zr0) / 2 + 0.5, (y * zr0) / 1.8 + 0.5)
			gl.Normal(x * zr0, y * zr0, z0)
			gl.Vertex(x * zr0, y * zr0, z0)
			gl.TexCoord((x * zr1) / 2 + 0.5, (y * zr1) / 1.8 + 0.5)
			gl.Normal(x * zr1, y * zr1, z1)
			gl.Vertex(x * zr1, y * zr1, z1)
		end
	
		gl.End()
	end
end

function win:key(e, k)
	if(e == "down") then
		if(k == 27) then
			self.fullscreen = not self.fullscreen
		end
	end
end

function win:init() 
	genMesh(1, dim, dim)
	sphereMesh:vertex(vertices)
	sphereMesh:index(indices)
	sphereMesh:normal(normals)
	sphereMesh:texcoord(texcoords)
end

function win:draw()

	frameTex.dim = self.dim
	frameTex:startcapture()

	gl.Enable(GL.MULTISAMPLE)
	gl.PushMatrix()
	
	gl.Rotate(now() * 10, 1, -1, .5)

---[[
	gl.DepthMask(GL.FALSE)
	gl.BlendFunc(GL.SRC_ALPHA, GL.ONE)
	gl.Enable(GL.BLEND)

	psShader:bind()
	blobTex:bind(1)

	gl.Begin(GL.POINTS)
		gl.Color(.3, .4, 1, .5) 
		gl.Vertex(-.3, 0, -.5)
		gl.Color(.7, .4, .3, .5) 
		gl.Vertex(0, .0, 0)
		gl.Color(.4, .9, .3, .5) 
		gl.Vertex(0.2, 0, 0.5)
		gl.Color(.2, 4, .7, .5) 
		gl.Vertex(-0.2, -0.3, -0.5)
	gl.End()

	blobTex:unbind(1)
	psShader:unbind()

	gl.Disable(GL.BLEND)
	gl.DepthMask(GL.TRUE)

--]]

---[[
	
	gl.DepthMask(GL.FALSE)
	gl.BlendFunc(GL.SRC_ALPHA, GL.ONE_MINUS_SRC_ALPHA)
	gl.Enable(GL.BLEND)
--	gl.Enable(GL.CULL_FACE)

--	gl.AlphaFunc(GL.GREATER, 0.4)
--	gl.Enable(GL.ALPHA_TEST)
	
	gl.PushMatrix()
	gl.Rotate(now() * 2, -1, 1, -.5)

	cutTex:bind()
--		gl.Color(1, 1, 1, 1) 
		gl.Color(0, 0, 0, 1) 
		sphereMesh:draw()
	cutTex:unbind()
	gl.PopMatrix()

--	gl.Disable(GL.ALPHA_TEST)

	gl.Disable(GL.BLEND)
	gl.DepthMask(GL.TRUE)

--]]

	gl.PopMatrix()
	gl.Disable(GL.MULTISAMPLE)

	frameTex:endcapture()
	
	gl.BlendFunc(GL.SRC_ALPHA, GL.ONE)
	gl.Enable(GL.BLEND)
	gl.DepthMask(GL.FALSE)	

	sketch.enter_ortho(0, 0, 1, 1)

---[[
	frameTex:bind(0)
	gl.Begin(GL.QUADS)
		gl.Color(1, 1, 1, 1)
		sketch.quad(0, 0, 1, 1)
	gl.End()
	frameTex:unbind(0)
--]]

---[[	
	lsShader:bind()
	frameTex:bind(0)
	gl.Begin(GL.QUADS)
		sketch.quad(0, 0, 1, 1)
	gl.End()
	frameTex:unbind(0)
	lsShader:unbind()
--]]

	sketch.leave_ortho()

	gl.DepthMask(GL.TRUE)	
	gl.Disable(GL.BLEND)
	
end